import numpy as np
np_load_old = np.load
np.load = lambda *a,**k: np_load_old(*a, allow_pickle=True, **k)
import pandas as pd
import os
import time
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader,TensorDataset
import random

os.chdir("/lustre1/hqzhu_pkuhpc/jtan/HoPhage_revise/data")


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(device)


def get_traindata(bs):
    labels = torch.from_numpy(np.load("HoPhage-G/allphage/Input_train_label-400-800.npy").astype('uint8'))
    input1 = torch.from_numpy(np.load("HoPhage-G/allphage/Input_train_seq-400-800.npy").astype('uint8'))
    input1 = input1.permute(0, 2, 1)
    input2 = torch.from_numpy(np.load("HoPhage-G/allphage/Input_train_01-400-800.npy").astype('uint8'))
    input2 = input2.permute(0, 2, 1)
    input12 =  torch.cat((input1,input2), dim=1)
    del input1,input2
    DICODON = torch.from_numpy(np.load("HoPhage-G/DICODON.npy"))
    MER = torch.from_numpy(np.load("HoPhage-G/MER.npy"))
    MER = MER.unsqueeze(1)
    host_taxa = pd.read_csv("HoPhage-G/host_taxa_1353.csv", keep_default_na=False)

    train_dataset = TensorDataset(input12, labels)
    train_loader = DataLoader(train_dataset, batch_size=bs, shuffle=True, pin_memory=True, num_workers=4)

    return train_loader,DICODON,MER,host_taxa


class Inception(nn.Module):
    def __init__(self, in_ch, out_ch1, mid_ch13, out_ch13, mid_ch15, out_ch15, out_ch_pool_conv):
        super(Inception, self).__init__()

        self.conv1 = nn.Sequential(
            nn.Conv1d(in_ch, out_ch1, kernel_size=1, stride=1),
            nn.ReLU())
        self.conv13 = nn.Sequential(
            nn.Conv1d(in_ch, mid_ch13, kernel_size=1, stride=1),
            nn.ReLU(),
            nn.Conv1d(mid_ch13, out_ch13, kernel_size=3, stride=1, padding=1),
            nn.ReLU())

        self.conv15 = nn.Sequential(
            nn.Conv1d(in_ch, mid_ch15, kernel_size=1, stride=1),
            nn.ReLU(),
            nn.Conv1d(mid_ch15, out_ch15, kernel_size=5, stride=1, padding=2),
            nn.ReLU())

        self.pool_conv1 = nn.Sequential(
            nn.MaxPool1d(3, stride=1, padding=1),
            nn.Conv1d(in_ch, out_ch_pool_conv, kernel_size=1, stride=1),
            nn.ReLU())

    def forward(self, inputs, train=False):
        conv1_out = self.conv1(inputs)
        conv13_out = self.conv13(inputs)
        conv15_out = self.conv15(inputs)
        pool_conv_out = self.pool_conv1(inputs)
        outputs = torch.cat([conv1_out, conv13_out, conv15_out, pool_conv_out], 1)  # depth-wise concat

        return outputs


class MyNet(nn.Module):
    def __init__(self):
        super(MyNet, self).__init__()

        self.layer12_0 = nn.Sequential(
            nn.Conv1d(6, 64, 7, 2, 3),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(3, 2, 1),
            nn.Conv1d(64, 64, 1),
            nn.ReLU(),
            nn.Conv1d(64, 192, 3, 1, 1),
            nn.BatchNorm1d(192),
            nn.ReLU(),
            nn.MaxPool1d(3, 2, 1)
        )
        # in_ch,out_ch_1,mid_ch_13,out_ch_13,mid_ch_15,out_ch_15,out_ch_pool_conv
        self.layer12_1 = nn.Sequential(
            Inception(192, 64, 96, 128, 16, 32, 32),
            Inception(256, 128, 128, 256, 32, 64, 64),
            # nn.AvgPool1d(3, 2, 1)
            nn.AdaptiveAvgPool1d(1)   
        )

        
        self.layer3_0 = nn.Sequential(
            nn.Conv1d(in_channels=64, out_channels=512, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(512),
            nn.ReLU(),
            # nn.AvgPool1d(kernel_size=3, stride=2, padding=1),
            nn.AdaptiveAvgPool1d(1)
        )

        self.layer4_0 = nn.Sequential(
            nn.Linear(1024,512),
            nn.ReLU(),
            nn.Dropout(0.3)   
        )

        self.layer5 = nn.Sequential(
            nn.Linear(512*3, 512),
            nn.ReLU(),
            nn.Dropout(0.3)  
        )

        self.output = nn.Linear(512, 2)
        # self.output = nn.Linear(512, 1)

    def forward(self, input12,input3,input4):
        out12 = self.layer12_0(input12)
        out12 = self.layer12_1(out12)

        out3 = self.layer3_0(input3)

        out4 = self.layer4_0(input4)

        out = torch.cat([out12.view(-1,512), out3.view(-1,512), out4.view(-1,512)], 1)
        out = self.layer5(out)
        out = self.output(out)

        # return F.softmax(out, dim=1)
        return out


def train(net, train_loader, DICODON, MER, epoch_num, epoch_sum, loss_fn, optimizer, scheduler, device):
    net.train()
    for epoch in range(epoch_num):
        total_loss = 0
        print_loss = 0
        i = 0
        for input12, labels in train_loader:
            input12 = input12.type(torch.FloatTensor)
            labels = labels.numpy()
            ind = np.where(labels == 1)[1]
            genus = host_taxa.genus[ind]
            genus = genus.drop_duplicates().values
            ind1 = host_taxa[host_taxa['genus'].isin(genus)].index
            family = host_taxa[host_taxa['genus'].isin(genus)].family
            family = family.drop_duplicates()
            temp = host_taxa[~host_taxa['family'].isin(family.values)].index
            ind2 = random.sample(list(temp), 4)

            input12 = input12.repeat_interleave(len(ind1) + len(ind2), dim=0)
            temp = torch.cat((DICODON[ind1], DICODON[ind2]), dim=0)
            input3 = temp.repeat(len(labels), 1, 1)
            input3 = input3.type(torch.FloatTensor)
            del temp
            temp = torch.cat((MER[ind1], MER[ind2]), dim=0)
            input4 = temp.repeat(len(labels), 1, 1)
            input4 = input4.type(torch.FloatTensor)
            del temp
            label = np.concatenate((labels[0][ind1], labels[0][ind2]))
            for j in range(1, len(labels)):
                label = np.concatenate((label, labels[j][ind1], labels[j][ind2]))
            label = torch.from_numpy(label)
            label = label.type(torch.LongTensor)

            input12, input3, input4, label = input12.to(device), input3.to(device), input4.to(
                device), label.to(device)
            # print(input1.size())
            # print(input2.size())
            # print(input3.size())
            # print(input4.size())
            # print(labels.size())

            # zero the parameter gradients
            optimizer.zero_grad()

            # forward + backward + optimize
            outputs = net(input12, input3, input4)
            loss = loss_fn(outputs, label)
            loss.backward()
            optimizer.step()

            print_loss += loss.item()
            if (i + 1) % 50000 == 0:
                print('batch = %d, loss = %.4f' % (i + 1, print_loss / 50000))
                print_loss = 0

            total_loss += loss.item()
            i += 1

        scheduler.step(loss)
        avg_loss = total_loss / i
        print("epoch %d average loss：%.4f" % (epoch, avg_loss))

    torch.save({
        'epoch': epoch_sum,
        'model_state_dict': net.state_dict(),
        'optimizer_state_dict': optimizer.state_dict(),
        'loss': loss,
        'scheduler':scheduler
    }, 'HoPhage-G/allphage/model_400-800_epoch'+str(epoch_sum)+'.pth')

    return net, optimizer, scheduler


if __name__ == '__main__':
    bs = 8  #batch size
    t0 = time.time()
    train_loader,DICODON,MER,host_taxa = get_traindata(bs)
    print('Time used for loading train data: {} sec'.format(time.time() - t0))

    net = MyNet()
    # summary(net.to(device), [(4, 2400), (2, 2400), (64, 64), (1, 1024)])
    net.to(device)
    # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # summary(net.to(device), [(4, 2400), (2, 2400), (64, 64), (1, 1024)])

    loss_fn = nn.CrossEntropyLoss()
    optimizer = optim.Adam(net.parameters(), lr=0.001, betas=(0.9, 0.99))
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.1, patience=2,
                                                           verbose=True, threshold=0.0001, threshold_mode='rel',
                                                           cooldown=0, min_lr=0, eps=1e-08)

    epoch_sum = 0
	
    epoch_num = 6
    epoch_sum += epoch_num
    net, optimizer, scheduler = train(net, train_loader, DICODON, MER, epoch_num, epoch_sum, loss_fn, optimizer, scheduler, device)

    epoch_num = 2
    epoch_sum += epoch_num
    net, optimizer, scheduler = train(net, train_loader, DICODON, MER, epoch_num, epoch_sum, loss_fn, optimizer,
                                      scheduler, device)

    epoch_num = 2
    epoch_sum += epoch_num
    net, optimizer, scheduler = train(net, train_loader, DICODON, MER, epoch_num, epoch_sum, loss_fn, optimizer,
                                      scheduler, device)

    epoch_num = 2
    epoch_sum += epoch_num
    net, optimizer, scheduler = train(net, train_loader, DICODON, MER, epoch_num, epoch_sum, loss_fn, optimizer, scheduler, device)

    # epoch_num = 3
    # epoch_sum += epoch_num
    # net, optimizer, scheduler = train(net, train_loader, DICODON, MER, epoch_num, epoch_sum, loss_fn, optimizer, scheduler, device)


    print('Finished')
